@@ -29,8 +29,8 @@
 #include "gspell-checker.h"
 #include "gspell-buffer-notifier.h"
 #include "gspell-text-buffer.h"
+#include "gspell-text-region.h"
 #include "gspell-utils.h"
-#include "gtktextregion.h"
 
 struct _GspellInlineCheckerTextBuffer
 {
@@ -47,8 +47,21 @@ struct _GspellInlineCheckerTextBuffer
 
 	GtkTextMark *mark_click;
 
-	GtkTextRegion *scan_region;
+	GspellTextRegion *scan_region;
 	guint timeout_id;
+
+	/* When a word is being typed, it should not be marked as misspelled,
+	 * because it would be annoying.
+	 */
+	guint check_current_word : 1;
+
+	/* If the unit test mode is enabled, there is no timeouts, and the whole
+	 * buffer is scanned synchronously.
+	 * The unit test mode tries to follow as most as possible the same code
+	 * paths as the real code paths, otherwise the unit tests would be
+	 * useless. As such, the function names reflect the real code paths.
+	 */
+	guint unit_test_mode : 1;
 };
 
 enum
@@ -57,8 +70,6 @@ enum
 	PROP_BUFFER,
 };
 
-#define ENABLE_DEBUG 0
-
 #define INLINE_CHECKER_TEXT_BUFFER_KEY	"GspellInlineCheckerTextBufferID"
 #define SUGGESTION_KEY			"GspellInlineSuggestionID"
 
@@ -101,7 +112,7 @@ check_word (GspellInlineCheckerTextBuffer *spell,
 	if (error != NULL)
 	{
 		g_warning ("Inline spell checker: %s", error->message);
-		g_error_free (error);
+		g_clear_error (&error);
 	}
 
 	if (!correctly_spelled)
@@ -213,30 +224,30 @@ get_visible_region (GtkTextView *view,
 
 /* A TextRegion can contain empty subregions. So checking the number of
  * subregions is not sufficient.
- * When calling gtk_text_region_add() with equal iters, the subregion is not
+ * When calling text_region_add() with equal iters, the subregion is not
  * added. But when a subregion becomes empty, due to text deletion, the
  * subregion is not removed from the TextRegion.
  */
 static gboolean
-is_text_region_empty (GtkTextRegion *region)
+is_text_region_empty (GspellTextRegion *region)
 {
-	GtkTextRegionIterator region_iter;
+	GspellTextRegionIterator region_iter;
 
 	if (region == NULL)
 	{
 		return TRUE;
 	}
 
-	gtk_text_region_get_iterator (region, &region_iter, 0);
+	_gspell_text_region_get_iterator (region, &region_iter, 0);
 
-	while (!gtk_text_region_iterator_is_end (&region_iter))
+	while (!_gspell_text_region_iterator_is_end (&region_iter))
 	{
 		GtkTextIter region_start;
 		GtkTextIter region_end;
 
-		if (!gtk_text_region_iterator_get_subregion (&region_iter,
-							     &region_start,
-							     &region_end))
+		if (!_gspell_text_region_iterator_get_subregion (&region_iter,
+								 &region_start,
+								 &region_end))
 		{
 			return TRUE;
 		}
@@ -246,95 +257,170 @@ is_text_region_empty (GtkTextRegion *region)
 			return FALSE;
 		}
 
-		gtk_text_region_iterator_next (&region_iter);
+		_gspell_text_region_iterator_next (&region_iter);
 	}
 
 	return TRUE;
 }
 
+/* Returns TRUE if there is a current word. */
+static gboolean
+get_current_word_boundaries (GtkTextBuffer *buffer,
+			     GtkTextIter   *current_word_start,
+			     GtkTextIter   *current_word_end)
+{
+	GtkTextIter insert_iter;
+	gboolean found = FALSE;
+
+	g_assert (current_word_start != NULL);
+	g_assert (current_word_end != NULL);
+
+	if (gtk_text_buffer_get_has_selection (buffer))
+	{
+		return FALSE;
+	}
+
+	gtk_text_buffer_get_iter_at_mark (buffer,
+					  &insert_iter,
+					  gtk_text_buffer_get_insert (buffer));
+
+	*current_word_start = insert_iter;
+	if (gtk_text_iter_ends_word (current_word_start) ||
+	    (gtk_text_iter_inside_word (current_word_start) &&
+	     !gtk_text_iter_starts_word (current_word_start)))
+	{
+		gtk_text_iter_backward_word_start (current_word_start);
+		found = TRUE;
+	}
+
+	*current_word_end = insert_iter;
+	if (gtk_text_iter_inside_word (current_word_end))
+	{
+		gtk_text_iter_forward_word_end (current_word_end);
+		found = TRUE;
+	}
+
+	return found;
+}
+
 static void
 check_visible_region_in_view (GspellInlineCheckerTextBuffer *spell,
 			      GtkTextView                   *view)
 {
 	GtkTextIter visible_start;
 	GtkTextIter visible_end;
-	GtkTextRegion *intersect;
-	GtkTextRegionIterator intersect_iter;
+	GspellTextRegion *intersect;
+	GspellTextRegionIterator intersect_iter;
 
 	if (spell->scan_region == NULL)
 	{
 		return;
 	}
 
-	get_visible_region (view, &visible_start, &visible_end);
+	if (view != NULL)
+	{
+		get_visible_region (view, &visible_start, &visible_end);
+	}
+	else
+	{
+		g_assert_true (spell->unit_test_mode);
+		gtk_text_buffer_get_bounds (spell->buffer, &visible_start, &visible_end);
+	}
 
-	intersect = gtk_text_region_intersect (spell->scan_region,
-					       &visible_start,
-					       &visible_end);
+	intersect = _gspell_text_region_intersect (spell->scan_region,
+						   &visible_start,
+						   &visible_end);
 
 	if (intersect == NULL)
 	{
 		return;
 	}
 
-	gtk_text_region_get_iterator (intersect, &intersect_iter, 0);
+	if (!spell->check_current_word)
+	{
+		GtkTextIter current_word_start;
+		GtkTextIter current_word_end;
 
-	while (!gtk_text_region_iterator_is_end (&intersect_iter))
+		if (get_current_word_boundaries (spell->buffer,
+						 &current_word_start,
+						 &current_word_end))
+		{
+			gtk_text_buffer_remove_tag (spell->buffer,
+						    spell->highlight_tag,
+						    &current_word_start,
+						    &current_word_end);
+
+			_gspell_text_region_subtract (intersect,
+						      &current_word_start,
+						      &current_word_end);
+		}
+	}
+
+	if (is_text_region_empty (intersect))
+	{
+		_gspell_text_region_destroy (intersect);
+		return;
+	}
+
+	_gspell_text_region_get_iterator (intersect, &intersect_iter, 0);
+
+	while (!_gspell_text_region_iterator_is_end (&intersect_iter))
 	{
 		GtkTextIter start;
 		GtkTextIter end;
 
-		if (!gtk_text_region_iterator_get_subregion (&intersect_iter, &start, &end))
+		if (!_gspell_text_region_iterator_get_subregion (&intersect_iter,
+								 &start,
+								 &end))
 		{
-			return;
+			break;
 		}
 
 		check_subregion (spell, &start, &end);
 
-		gtk_text_region_subtract (spell->scan_region, &start, &end);
+		_gspell_text_region_subtract (spell->scan_region, &start, &end);
 
-		gtk_text_region_iterator_next (&intersect_iter);
+		_gspell_text_region_iterator_next (&intersect_iter);
 	}
 
-	gtk_text_region_destroy (intersect);
+	_gspell_text_region_destroy (intersect);
 
 	if (is_text_region_empty (spell->scan_region))
 	{
-		gtk_text_region_destroy (spell->scan_region);
+		_gspell_text_region_destroy (spell->scan_region);
 		spell->scan_region = NULL;
 	}
+
+	if (view != NULL)
+	{
+		/* FIXME properly. Workaround for bug in GtkTextView:
+		 * https://bugzilla.gnome.org/show_bug.cgi?id=612772
+		 */
+		gtk_widget_queue_draw (GTK_WIDGET (view));
+	}
 }
 
 static void
 check_visible_region (GspellInlineCheckerTextBuffer *spell)
 {
 	GSList *l;
-#if ENABLE_DEBUG
-	GTimer *timer;
-#endif
 
 	if (spell->scan_region == NULL)
 	{
 		return;
 	}
 
-#if ENABLE_DEBUG
-	timer = g_timer_new ();
-#endif
+	if (spell->unit_test_mode)
+	{
+		check_visible_region_in_view (spell, NULL);
+		return;
+	}
 
 	for (l = spell->views; l != NULL; l = l->next)
 	{
 		GtkTextView *view = GTK_TEXT_VIEW (l->data);
 		check_visible_region_in_view (spell, view);
 	}
-
-#if ENABLE_DEBUG
-	g_print ("%s() executed in %lf seconds\n",
-		 G_STRFUNC,
-		 g_timer_elapsed (timer, NULL));
-
-	g_timer_destroy (timer);
-#endif
 }
 
 static gboolean
@@ -353,11 +439,19 @@ install_timeout (GspellInlineCheckerTextBuffer *spell,
 	if (spell->timeout_id != 0)
 	{
 		g_source_remove (spell->timeout_id);
+		spell->timeout_id = 0;
 	}
 
-	spell->timeout_id = g_timeout_add (duration,
-					   (GSourceFunc) timeout_cb,
-					   spell);
+	if (spell->unit_test_mode)
+	{
+		timeout_cb (spell);
+	}
+	else
+	{
+		spell->timeout_id = g_timeout_add (duration,
+						   (GSourceFunc) timeout_cb,
+						   spell);
+	}
 }
 
 static void
@@ -367,10 +461,10 @@ add_subregion_to_scan (GspellInlineCheckerTextBuffer *spell,
 {
 	if (spell->scan_region == NULL)
 	{
-		spell->scan_region = gtk_text_region_new (spell->buffer);
+		spell->scan_region = _gspell_text_region_new (spell->buffer);
 	}
 
-	gtk_text_region_add (spell->scan_region, start, end);
+	_gspell_text_region_add (spell->scan_region, start, end);
 }
 
 static void
@@ -381,6 +475,7 @@ recheck_all (GspellInlineCheckerTextBuffer *spell)
 
 	gtk_text_buffer_get_bounds (spell->buffer, &start, &end);
 
+	spell->check_current_word = TRUE;
 	add_subregion_to_scan (spell, &start, &end);
 	check_visible_region (spell);
 }
@@ -398,6 +493,21 @@ insert_text_after_cb (GtkTextBuffer                 *buffer,
 	start = end = *location;
 	gtk_text_iter_backward_chars (&start, g_utf8_strlen (text, length));
 
+	/* Include neighbor words */
+
+	if (gtk_text_iter_ends_word (&start) ||
+	    (gtk_text_iter_inside_word (&start) &&
+	     !gtk_text_iter_starts_word (&start)))
+	{
+		gtk_text_iter_backward_word_start (&start);
+	}
+
+	if (gtk_text_iter_inside_word (&end))
+	{
+		gtk_text_iter_forward_word_end (&end);
+	}
+
+	spell->check_current_word = FALSE;
 	add_subregion_to_scan (spell, &start, &end);
 	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
 }
@@ -429,10 +539,24 @@ delete_range_after_cb (GtkTextBuffer                 *buffer,
 		gtk_text_iter_forward_word_end (&end_adjusted);
 	}
 
+	spell->check_current_word = FALSE;
 	add_subregion_to_scan (spell, &start_adjusted, &end_adjusted);
 	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
 }
 
+static void
+mark_set_after_cb (GtkTextBuffer                 *buffer,
+		   GtkTextIter                   *location,
+		   GtkTextMark                   *mark,
+		   GspellInlineCheckerTextBuffer *spell)
+{
+	if (mark == gtk_text_buffer_get_insert (buffer))
+	{
+		spell->check_current_word = TRUE;
+		install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
+	}
+}
+
 static gboolean
 get_word_extents_at_click_position (GspellInlineCheckerTextBuffer *spell,
 				    GtkTextIter                   *start,
@@ -942,6 +1066,12 @@ set_buffer (GspellInlineCheckerTextBuffer *spell,
 				 spell,
 				 G_CONNECT_AFTER);
 
+	g_signal_connect_object (buffer,
+				 "mark-set",
+				 G_CALLBACK (mark_set_after_cb),
+				 spell,
+				 G_CONNECT_AFTER);
+
 	g_signal_connect_object (buffer,
 				 "apply-tag",
 				 G_CALLBACK (apply_or_remove_tag_cb),
@@ -1080,7 +1210,7 @@ _gspell_inline_checker_text_buffer_dispose (GObject *object)
 
 	if (spell->scan_region != NULL)
 	{
-		gtk_text_region_destroy (spell->scan_region);
+		_gspell_text_region_destroy (spell->scan_region);
 		spell->scan_region = NULL;
 	}
 
@@ -1116,6 +1246,7 @@ _gspell_inline_checker_text_buffer_class_init (GspellInlineCheckerTextBufferClas
 static void
 _gspell_inline_checker_text_buffer_init (GspellInlineCheckerTextBuffer *spell)
 {
+	spell->check_current_word = TRUE;
 }
 
 GspellInlineCheckerTextBuffer *
@@ -1179,6 +1310,7 @@ _gspell_inline_checker_text_buffer_attach_view (GspellInlineCheckerTextBuffer *s
 
 	spell->views = g_slist_prepend (spell->views, view);
 
+	spell->check_current_word = TRUE;
 	check_visible_region_in_view (spell, view);
 }
 
@@ -1202,4 +1334,30 @@ _gspell_inline_checker_text_buffer_detach_view (GspellInlineCheckerTextBuffer *s
 	spell->views = g_slist_remove (spell->views, view);
 }
 
+void
+_gspell_inline_checker_text_buffer_set_unit_test_mode (GspellInlineCheckerTextBuffer *spell,
+						       gboolean                       unit_test_mode)
+{
+	g_return_if_fail (GSPELL_IS_INLINE_CHECKER_TEXT_BUFFER (spell));
+
+	spell->unit_test_mode = unit_test_mode != FALSE;
+
+	if (spell->timeout_id != 0)
+	{
+		g_source_remove (spell->timeout_id);
+		spell->timeout_id = 0;
+		timeout_cb (spell);
+	}
+
+	check_visible_region (spell);
+}
+
+GtkTextTag *
+_gspell_inline_checker_text_buffer_get_highlight_tag (GspellInlineCheckerTextBuffer *spell)
+{
+	g_return_val_if_fail (GSPELL_IS_INLINE_CHECKER_TEXT_BUFFER (spell), NULL);
+
+	return spell->highlight_tag;
+}
+
 /* ex:set ts=8 noet: */
