@@ -24,13 +24,14 @@
 
 #include <gspell/gspell-checker.h>
 #include <gspell/gspell-checker-dialog.h>
-#include <gspell/gspell-inline-checker-gtv.h>
+#include <gspell/gspell-inline-checker-text-view.h>
 #include <gspell/gspell-language.h>
 #include <gspell/gspell-language-chooser.h>
 #include <gspell/gspell-language-chooser-button.h>
 #include <gspell/gspell-language-chooser-dialog.h>
 #include <gspell/gspell-navigator.h>
-#include <gspell/gspell-navigator-gtv.h>
+#include <gspell/gspell-navigator-text-view.h>
+#include <gspell/gspell-text-buffer.h>
 
 #undef __GSPELL_H_INSIDE__
 
